"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const BlinkDiff = require("blink-diff");
const PngJsImage = require("pngjs-image");
const image_options_1 = require("./image-options");
const log_image_type_1 = require("./enums/log-image-type");
const utils_1 = require("./utils");
const fs_1 = require("fs");
const path_1 = require("path");
class ImageHelper {
    constructor(_args, _driver) {
        this._args = _args;
        this._driver = _driver;
        this._imagesResults = new Map();
        this._options = {
            timeOutSeconds: 2,
            tolerance: 0,
            toleranceType: image_options_1.ImageOptions.pixel,
            waitOnCreatingInitialSnapshot: 2000,
            preserveImageName: false,
        };
        this.options.cropRectangele = (this._args.appiumCaps && this._args.appiumCaps.viewportRect) || this._args.device.viewportRect;
    }
    get options() {
        return this._options;
    }
    set options(options) {
        this._options = this.extendOptions(options);
    }
    set testName(testName) {
        this._testName = testName;
    }
    get testName() {
        return this._testName;
    }
    get imageComppareOptions() {
        this.extendOptions(this._options);
        return this._options;
    }
    set imageComppareOptions(imageComppareOptions) {
        this._options = this.extendOptions(imageComppareOptions);
    }
    compareScreen(options) {
        return __awaiter(this, void 0, void 0, function* () {
            options = this.extendOptions(options);
            options.imageName = this.increaseImageName(options.imageName || this._testName);
            const result = yield this.compare(options);
            this._imagesResults.set(options.imageName, result);
            return result;
        });
    }
    compareElement(element, options) {
        return __awaiter(this, void 0, void 0, function* () {
            options = this.extendOptions(options);
            options.imageName = this.increaseImageName(options.imageName || this._testName);
            const cropRectangele = yield element.getRectangle();
            const result = yield this.compareRectangle(cropRectangele, options);
            return result;
        });
    }
    compareRectangle(cropRectangle, options) {
        return __awaiter(this, void 0, void 0, function* () {
            options = this.extendOptions(options);
            options.imageName = this.increaseImageName(options.imageName || this._testName);
            options.cropRectangele = cropRectangle;
            const result = yield this.compare(options);
            this._imagesResults.set(options.imageName, result);
            return result;
        });
    }
    hasImageComparisonPassed() {
        let shouldFailTest = true;
        console.log();
        this._imagesResults.forEach((v, k, map) => {
            if (!this._imagesResults.get(k)) {
                shouldFailTest = false;
                this._driver.testReporterLog(`Image comparison for image ${k} has failed!`);
                utils_1.logError(`Image comparison for image ${k} has failed`);
            }
        });
        this.reset();
        return shouldFailTest;
    }
    reset() {
        this._imagesResults.clear();
        this.testName = undefined;
    }
    increaseImageName(imageName) {
        if (!imageName) {
            utils_1.logError(`Missing image name!`);
            utils_1.logError(`Consider to set
            drive.imageHelper.testName
            dirver.imageHelper.options.imageName
    `);
            throw new Error(`Missing image name!`);
        }
        if (this._imagesResults.size > 0) {
            const images = new Array();
            this._imagesResults.forEach((v, k, map) => {
                if (k.includes(imageName)) {
                    images.push(k);
                }
            });
            images.sort((a, b) => {
                const aNumber = +/\d+$/.exec(a);
                const bNumber = +/\d+$/.exec(b);
                return bNumber - aNumber;
            });
            if (images.length > 0) {
                const lastImage = images[0];
                const number = /\d+$/.test(lastImage) ? +`${/\d+$/.exec(lastImage)}` + 1 : `2`;
                imageName = `${imageName}_${number}`;
            }
        }
        return imageName;
    }
    extendOptions(options) {
        options = options || {};
        const clipRectangele = this.imageCropRect;
        Object.getOwnPropertyNames(this.options).forEach(prop => {
            if (!options[prop]) {
                options[prop] = this.options[prop];
            }
        });
        if (!options.cropRectangele) {
            Object.getOwnPropertyNames(clipRectangele).forEach(prop => {
                options.cropRectangele[prop] = clipRectangele[prop];
            });
            this.imageCropRect = options.cropRectangele;
        }
        return options;
    }
    get imageCropRect() {
        return this._imageCropRect || this.options.cropRectangele;
    }
    set imageCropRect(clipRectangele) {
        this._imageCropRect = clipRectangele;
    }
    get blockOutAreas() {
        return this._blockOutAreas;
    }
    set blockOutAreas(rectangles) {
        this._blockOutAreas = rectangles;
    }
    imageOutputLimit() {
        return image_options_1.ImageOptions.outputAll;
    }
    thresholdType() {
        return image_options_1.ImageOptions.percent;
    }
    threshold(thresholdType) {
        if (thresholdType == image_options_1.ImageOptions.percent) {
            return 0.01; // 0.01 = 1 percent; 500 percent
        }
        else {
            return 10; // 10 pixels
        }
    }
    delta() {
        return 20;
    }
    getExpectedImagePath(imageName) {
        let pathExpectedImage = utils_1.resolvePath(this._args.storageByDeviceName, imageName);
        return pathExpectedImage;
    }
    compare(options) {
        return __awaiter(this, void 0, void 0, function* () {
            let imageName = utils_1.addExt(options.imageName, ImageHelper.pngFileExt);
            const pathExpectedImage = this.getExpectedImagePath(imageName);
            // First time capture
            if (!fs_1.existsSync(pathExpectedImage)) {
                const pathActualImage = utils_1.resolvePath(this._args.storageByDeviceName, this.options.preserveImageName ? imageName : imageName.replace(".", "_actual."));
                if (this.options.waitOnCreatingInitialSnapshot > 0) {
                    yield this._driver.wait(this.options.waitOnCreatingInitialSnapshot);
                }
                yield this._driver.saveScreenshot(pathActualImage);
                // if (options.cropRectangele) {
                //     await this.clipRectangleImage(options.cropRectangele, pathActualImage);
                // }
                const pathActualImageToReportsFolder = utils_1.resolvePath(this._args.reportsPath, path_1.basename(pathActualImage));
                utils_1.copy(pathActualImage, pathActualImageToReportsFolder, false);
                if (this.options.preserveImageName) {
                    utils_1.logWarn(`New image ${path_1.basename(pathActualImage)} is saved to storage ${this._args.storageByDeviceName}.`, pathExpectedImage);
                }
                else {
                    utils_1.logWarn("Remove the 'actual' suffix to continue using the image as expected one ", pathExpectedImage);
                }
                this._args.testReporterLog(path_1.basename(pathActualImage).replace(/\.\w{3,3}$/ig, ""));
                this._args.testReporterLog(path_1.join(this._args.reportsPath, path_1.basename(pathActualImage)));
                return false;
            }
            // Compare
            let pathActualImage = yield this._driver.saveScreenshot(utils_1.resolvePath(this._args.reportsPath, imageName.replace(".", "_actual.")));
            // if (options.cropRectangele) {
            //     await this.clipRectangleImage(options.cropRectangele, pathActualImage);
            // }
            const pathDiffImage = pathActualImage.replace("actual", "diff");
            // await this.prepareImageToCompare(pathActualImage, options.cropRectangele);
            let result = yield this.compareImages(pathActualImage, pathExpectedImage, pathDiffImage, options.tolerance, options.toleranceType);
            // Iterate
            if (!result) {
                const eventStartTime = Date.now().valueOf();
                let counter = 1;
                options.timeOutSeconds *= 1000;
                while ((Date.now().valueOf() - eventStartTime) <= options.timeOutSeconds && !result) {
                    const pathActualImageConter = utils_1.resolvePath(this._args.reportsPath, imageName.replace(".", "_actual_" + counter + "."));
                    pathActualImage = yield this._driver.saveScreenshot(pathActualImageConter);
                    // if (options.cropRectangele) {
                    //     await this.clipRectangleImage(options.cropRectangele, pathActualImage);
                    // }
                    // await this.prepareImageToCompare(pathActualImage, this.imageCropRect);
                    result = yield this.compareImages(pathActualImage, pathExpectedImage, pathDiffImage, options.tolerance, options.toleranceType);
                    if (!result && utils_1.checkImageLogType(this._args.testReporter, log_image_type_1.LogImageType.everyImage)) {
                        this._args.testReporterLog(`Actual image: ${path_1.basename(pathActualImage).replace(/\.\w{3,3}$/ig, "")}`);
                        this._args.testReporterLog(path_1.join(this._args.reportsPath, path_1.basename(pathActualImage)));
                    }
                    counter++;
                }
                if (!result && !utils_1.checkImageLogType(this._args.testReporter, log_image_type_1.LogImageType.everyImage)) {
                    this._args.testReporterLog(`${path_1.basename(pathDiffImage).replace(/\.\w{3,3}$/ig, "")}`);
                    this._args.testReporterLog(path_1.join(this._args.reportsPath, path_1.basename(pathDiffImage)));
                    this._args.testReporterLog(`Actual image: ${path_1.basename(pathActualImage).replace(/\.\w{3,3}$/ig, "")}`);
                    this._args.testReporterLog(path_1.join(this._args.reportsPath, path_1.basename(pathActualImage)));
                }
            }
            else {
                if (fs_1.existsSync(pathDiffImage)) {
                    fs_1.unlinkSync(pathDiffImage);
                }
                if (fs_1.existsSync(pathActualImage)) {
                    fs_1.unlinkSync(pathActualImage);
                }
            }
            return result;
        });
    }
    // public async prepareImageToCompare(filePath: string, rect: IRectangle) {
    //     if (rect) {
    //         await this.clipRectangleImage(rect, filePath);
    //         const rectToCrop = { left: 0, top: 0, width: undefined, height: undefined };
    //         this.imageCropRect = rectToCrop;
    //     } else {
    //         this.imageCropRect = ImageHelper.cropImageDefault(this._args);
    //     }
    // }
    runDiff(diffOptions, diffImage) {
        var that = this;
        return new Promise((resolve, reject) => {
            diffOptions.run(function (error, result) {
                if (error) {
                    throw error;
                }
                else {
                    let message;
                    let resultCode = diffOptions.hasPassed(result.code);
                    if (resultCode) {
                        message = "Screen compare passed!";
                        console.log(message);
                        console.log('Found ' + result.differences + ' differences.');
                        return resolve(true);
                    }
                    else {
                        message = `Screen compare failed! Found ${result.differences} differences.\n`;
                        console.log(message);
                        if (Object.getOwnPropertyNames(that._args.testReporter).length > 0 && that._args.testReporter.logImageTypes && that._args.testReporter.logImageTypes.indexOf(log_image_type_1.LogImageType.everyImage) > -1) {
                            that._args.testReporterLog(`${path_1.basename(diffImage)}\n\r${message}`);
                            that._args.testReporterLog(diffImage);
                        }
                        return resolve(false);
                    }
                }
            });
        });
    }
    compareImages(actual, expected, output, valueThreshold = this.threshold(this.thresholdType()), typeThreshold = this.thresholdType()) {
        const clipRect = {
            x: this.imageCropRect.left,
            y: this.imageCropRect.top,
            width: this.imageCropRect.width,
            height: this.imageCropRect.height
        };
        const diff = new BlinkDiff({
            imageAPath: actual,
            imageBPath: expected,
            imageOutputPath: output,
            imageOutputLimit: this.imageOutputLimit(),
            thresholdType: typeThreshold,
            threshold: valueThreshold,
            delta: this.delta(),
            cropImageA: clipRect,
            cropImageB: clipRect,
            blockOut: this._blockOutAreas,
            verbose: this._args.verbose,
        });
        if (typeThreshold == image_options_1.ImageOptions.percent) {
            valueThreshold = Math.floor(valueThreshold * 100);
            console.log(`Using ${valueThreshold}\ ${typeThreshold} tolerance`);
        }
        else {
            console.log(`Using ${valueThreshold} tolerance`);
        }
        const result = this.runDiff(diff, output);
        this._blockOutAreas = undefined;
        return result;
    }
    clipRectangleImage(rect, path) {
        return __awaiter(this, void 0, void 0, function* () {
            let imageToClip;
            imageToClip = yield this.readImage(path);
            imageToClip.clip(rect.left, rect.top, rect.width, rect.height);
            return new Promise((resolve, reject) => {
                imageToClip.writeImage(path, (err) => {
                    if (err) {
                        return reject(err);
                    }
                    return resolve();
                });
            });
        });
    }
    readImage(path) {
        return new Promise((resolve, reject) => {
            PngJsImage.readImage(path, (err, image) => {
                if (err) {
                    return reject(err);
                }
                return resolve(image);
            });
        });
    }
}
ImageHelper.pngFileExt = '.png';
exports.ImageHelper = ImageHelper;
//# sourceMappingURL=image-helper.js.map